/******************************************************************************
 * @file     main.c
 * @brief
 *           Use ACMP to trigger EPWM brake, user can choice edge detect 
 *           or level detect by define.
 * @note
 * Copyright (C) 2019 Nuvoton Technology Corp. All rights reserved.
 *****************************************************************************/
#include "stdio.h"
#include "NuMicro.h"

/*---------------------------------------------------------------------------*/
/* Definitions                                                               */
/*---------------------------------------------------------------------------*/
//#define USE_EDGE_DETECT
#define USE_LEVEL_DETECT
/*---------------------------------------------------------------------------*/
/* Global variables                                                          */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Functions                                                                 */
/*---------------------------------------------------------------------------*/
void SYS_Init(void)
{
    /* Enable HXT clock (external XTAL 12MHz) */
    CLK_EnableXtalRC(CLK_PWRCTL_HXTEN_Msk);

    /* Wait for HXT clock ready */
    CLK_WaitClockReady(CLK_STATUS_HXTSTB_Msk);

    /* Set core clock as PLL_CLOCK from PLL */
    CLK_SetCoreClock(FREQ_192MHZ);
    
    /* Set PCLK0 = PCLK1 = HCLK/2 */
    CLK->PCLKDIV = (CLK_PCLKDIV_PCLK0DIV2 | CLK_PCLKDIV_PCLK1DIV2); // PCLK divider set 2

    /* Enable module clock */
    CLK_EnableModuleClock(UART0_MODULE);
    CLK_EnableModuleClock(EPWM0_MODULE);
    CLK_EnableModuleClock(ACMP01_MODULE);
    CLK_EnableModuleClock(DAC_MODULE);
    CLK_EnableModuleClock(TMR0_MODULE);

    /* Set module clock */
    CLK_SetModuleClock(UART0_MODULE, CLK_CLKSEL1_UART0SEL_HXT, CLK_CLKDIV0_UART0(1));
    CLK_SetModuleClock(EPWM0_MODULE, CLK_CLKSEL2_EPWM0SEL_PLL, (uint32_t)NULL);
    CLK_SetModuleClock(TMR0_MODULE, CLK_CLKSEL1_TMR0SEL_HXT, (uint32_t)NULL);
}

void UART0_Init()
{
    /* Set PB.12 for UART0_RXD, PB.13 for UART0_TXD */
    SYS->GPB_MFPH &= ~(SYS_GPB_MFPH_PB12MFP_Msk | SYS_GPB_MFPH_PB13MFP_Msk);
    SYS->GPB_MFPH |= (SYS_GPB_MFPH_PB12MFP_UART0_RXD | SYS_GPB_MFPH_PB13MFP_UART0_TXD);

    /* Configure UART0 and set UART0 baud rate */
    UART_Open(UART0, 115200);
}

void epwm_init(void)
{
    /* Set PA.5 as EPWM0_CH0 */
    SYS->GPA_MFPL &= ~(SYS_GPA_MFPL_PA5MFP_Msk);
    SYS->GPA_MFPL |= (SYS_GPA_MFPL_PA5MFP_EPWM0_CH0);
	
    /* Config EPWM0_CH0 output 1KHz, duty 50% waveform */
    EPWM_ConfigOutputChannel(EPWM0, 0, 1000, 50);
    EPWM_EnableOutput(EPWM0, 0x01);
    EPWM_Start(EPWM0, 0x01);

    /* Unlock protected registers */
    SYS_UnlockReg();
    
#ifdef USE_EDGE_DETECT
	/* Set EPWM0_0 brake source from ACMP0, and use edge detect */
    EPWM_EnableFaultBrake(EPWM0, EPWM_CH_0_MASK, 0, EPWM_FB_EDGE_ACMP0);
    
    /* Enbale EPWM0 edge-detect brake interrupt */
    EPWM_EnableFaultBrakeInt(EPWM0, EPWM_FB_EDGE);
#endif
    
#ifdef USE_LEVEL_DETECT
	/* Set EPWM0_0 brake source from ACMP0, and use level detect */
    EPWM_EnableFaultBrake(EPWM0, EPWM_CH_0_MASK, 0, EPWM_FB_LEVEL_ACMP0);
    
    /* Enbale EPWM0 level-detect brake interrupt */
    EPWM_EnableFaultBrakeInt(EPWM0, EPWM_FB_LEVEL);
#endif
    
    /* Lock protected registers */
    SYS_LockReg();
    
    /* Enable BRAKE0 interrupt */
    NVIC_EnableIRQ(BRAKE0_IRQn);
}

void gpio_init(void)
{
    /* Set PB.3 as GPIO */
    SYS->GPB_MFPL &= ~(SYS_GPB_MFPL_PB3MFP_Msk);
    SYS->GPB_MFPL |= (SYS_GPB_MFPL_PB3MFP_GPIO);
	
    /* Set PB.3 to output mode */
    GPIO_SetMode(PB, BIT3, GPIO_MODE_OUTPUT);
    
    PB3 = 0;
}

void acmp_init(void)
{
    /* Set PB.2 as ACMP0_P1 */
    SYS->GPB_MFPL &= ~(SYS_GPB_MFPL_PB2MFP_Msk);
    SYS->GPB_MFPL |= (SYS_GPB_MFPL_PB2MFP_ACMP0_P1);

    /* Disable digital input path of analog pin ACMP0_P1 to prevent leakage */
    GPIO_DISABLE_DIGITAL_PATH(PB, (1ul << 2));
    
    /* Open ACMP0 and set negative source from CRV */
    ACMP_Open(ACMP01, 0, ACMP_CTL_NEGSEL_CRV, ACMP_CTL_HYSTERESIS_DISABLE);
    
    /* Set CRV voltage source from AVDD */
    ACMP_SELECT_CRV_SRC(ACMP01, ACMP_VREF_CRVSSEL_VDDA);
    
    /* Set CRV level = 6
     * CRV = AVDD x (1/6 + level /24) = 3.3 x (1/6 + 6/24) = 1.375 V
     */
    ACMP_CRV_SEL(ACMP01, 6);    
    
    /* Select positive input from ACMP0_P1 */
    ACMP_SELECT_P(ACMP01, 0, ACMP_CTL_POSSEL_P1);

    /* Enbale ACMP Interrupt */
    ACMP_ENABLE_INT(ACMP01, 0);
    NVIC_EnableIRQ(ACMP01_IRQn);
}

void timer_init(void)
{
    /* Open timer */
    TIMER_Open(TIMER0, TIMER_PERIODIC_MODE, 2);
	
    /* Enable timer interrupt */
    TIMER_EnableInt(TIMER0);
    NVIC_EnableIRQ(TMR0_IRQn);
	
    /* Let timer start counting */
    TIMER_Start(TIMER0);
}

int32_t main(void)
{
    /* Unlock protected registers */
    SYS_UnlockReg();

    /* Init System, peripheral clock and multi-function I/O */
    SYS_Init();

    /* Lock protected registers */
    SYS_LockReg();

    /* Init UART0 for printf and test */
    UART0_Init();

    /* Init EPWM output */
    epwm_init();
    
    /* Init GPIO output */
    gpio_init();
    
    /* Init ACMP */
    acmp_init();
    
    /* Init Timer */
    timer_init();
    
    while(1);
}

void BRAKE0_IRQHandler(void)
{
#ifdef USE_EDGE_DETECT
	if (EPWM_GetFaultBrakeIntFlag(EPWM0, EPWM_FB_EDGE) == 1) {
		/* Unlock protected registers */
		SYS_UnlockReg();

		/* Clear PWM0 brake interrupt flag */
		EPWM_ClearFaultBrakeIntFlag(EPWM0, EPWM_FB_EDGE);
		
		printf("!!! Edge Brake happened !!!\n");

		/* Lock protected registers */
		SYS_LockReg();
	}
#endif
    
#ifdef USE_LEVEL_DETECT
	if (EPWM_GetFaultBrakeIntFlag(EPWM0, EPWM_FB_LEVEL) == 1) {
		/* Unlock protected registers */
		SYS_UnlockReg();
        
		/* Clear PWM0 brake interrupt flag */
		EPWM_ClearFaultBrakeIntFlag(EPWM0, EPWM_FB_LEVEL);
		
		printf("!!! Level Brake happened !!!\n");

		/* Lock protected registers */
		SYS_LockReg();
	}
#endif
}

void ACMP01_IRQHandler(void)
{
    /* Check Comparator 0 Output Status */
    if(ACMP_GET_OUTPUT(ACMP01, 0)) {
        printf("ACMP0_P0 voltage > CRV Voltage\n");
    } else {
        printf("ACMP0_P0 voltage <= CRV Voltage\n");	
    }
    
    /* Clear ACMP 0 interrupt flag */
    ACMP_CLR_INT_FLAG(ACMP01, 0);
}

void TMR0_IRQHandler(void)
{
    if (TIMER_GetIntFlag(TIMER0) == 1) {
        PB3 ^= 1;
        /* Clear Timer0 time-out interrupt flag */
        TIMER_ClearIntFlag(TIMER0);
    }
}


/*** (C) COPYRIGHT 2019 Nuvoton Technology Corp. ***/
